/*
This file is part of the MusiKernel project, Copyright MusiKernel Team

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*/

#ifndef EUPHORIA_PORTS_H
#define	EUPHORIA_PORTS_H

//Delimits the file string sent with configure().  Also used in the file saving format
#define EUPHORIA_FILES_STRING_DELIMITER '|'
//When used in place of "|", it tells the sampler to load the sample even if it's already been loaded once.

/*Provide an arbitrary maximum number of samples the user can load*/
#define EUPHORIA_MAX_SAMPLE_COUNT 100

//Total number of LFOs, ADSRs, other envelopes, etc...  Used for the PolyFX mod matrix
#define EUPHORIA_MODULATOR_COUNT 6
//How many modular PolyFX
#define EUPHORIA_MODULAR_POLYFX_COUNT 4
//How many ports per PolyFX, 3 knobs and a combobox
#define EUPHORIA_PORTS_PER_MOD_EFFECT 4
//How many knobs per PolyFX, 3 knobs
#define EUPHORIA_CONTROLS_PER_MOD_EFFECT 3

//The number of mono_fx groups
#define EUPHORIA_MONO_FX_GROUPS_COUNT EUPHORIA_MAX_SAMPLE_COUNT
#define EUPHORIA_MONO_FX_COUNT 4

#define EUPHORIA_LABEL "Euphoria"

#define EUPHORIA_FIRST_CONTROL_PORT 3

#define EUPHORIA_ATTACK  3
#define EUPHORIA_DECAY   4
#define EUPHORIA_SUSTAIN 5
#define EUPHORIA_RELEASE 6
#define EUPHORIA_FILTER_ATTACK  7
#define EUPHORIA_FILTER_DECAY   8
#define EUPHORIA_FILTER_SUSTAIN 9
#define EUPHORIA_FILTER_RELEASE 10
#define EUPHORIA_LFO_PITCH 11
#define EUPHORIA_MASTER_VOLUME 12
#define EUPHORIA_MASTER_GLIDE 13
#define EUPHORIA_MASTER_PITCHBEND_AMT 14
#define EUPHORIA_PITCH_ENV_TIME 15
#define EUPHORIA_LFO_FREQ 16
#define EUPHORIA_LFO_TYPE 17
//From Modulex
#define EUPHORIA_FX0_KNOB0  18
#define EUPHORIA_FX0_KNOB1 19
#define EUPHORIA_FX0_KNOB2  20
#define EUPHORIA_FX0_COMBOBOX 21
#define EUPHORIA_FX1_KNOB0  22
#define EUPHORIA_FX1_KNOB1  23
#define EUPHORIA_FX1_KNOB2  24
#define EUPHORIA_FX1_COMBOBOX 25
#define EUPHORIA_FX2_KNOB0  26
#define EUPHORIA_FX2_KNOB1  27
#define EUPHORIA_FX2_KNOB2  28
#define EUPHORIA_FX2_COMBOBOX 29
#define EUPHORIA_FX3_KNOB0  30
#define EUPHORIA_FX3_KNOB1  31
#define EUPHORIA_FX3_KNOB2  32
#define EUPHORIA_FX3_COMBOBOX 33
//PolyFX Mod Matrix
#define EUPHORIA_PFXMATRIX_FIRST_PORT 34

#define EUPHORIA_PFXMATRIX_GRP0DST0SRC0CTRL0  34
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC0CTRL1  35
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC0CTRL2  36
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC1CTRL0  37
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC1CTRL1  38
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC1CTRL2  39
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC2CTRL0  40
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC2CTRL1  41
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC2CTRL2  42
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC3CTRL0  43
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC3CTRL1  44
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC3CTRL2  45
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC0CTRL0  46
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC0CTRL1  47
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC0CTRL2  48
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC1CTRL0  49
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC1CTRL1  50
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC1CTRL2  51
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC2CTRL0  52
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC2CTRL1  53
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC2CTRL2  54
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC3CTRL0  55
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC3CTRL1  56
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC3CTRL2  57
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC0CTRL0  58
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC0CTRL1  59
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC0CTRL2  60
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC1CTRL0  61
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC1CTRL1  62
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC1CTRL2  63
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC2CTRL0  64
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC2CTRL1  65
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC2CTRL2  66
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC3CTRL0  67
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC3CTRL1  68
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC3CTRL2  69
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC0CTRL0  70
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC0CTRL1  71
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC0CTRL2  72
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC1CTRL0  73
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC1CTRL1  74
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC1CTRL2  75
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC2CTRL0  76
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC2CTRL1  77
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC2CTRL2  78
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC3CTRL0  79
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC3CTRL1  80
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC3CTRL2  81


#define EUPHORIA_PFXMATRIX_GRP0DST0SRC4CTRL0  82
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC4CTRL1  83
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC4CTRL2  84
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC4CTRL0  85
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC4CTRL1  86
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC4CTRL2  87
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC4CTRL0  88
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC4CTRL1  89
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC4CTRL2  90
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC4CTRL0  91
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC4CTRL1  92
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC4CTRL2  93

#define EUPHORIA_PFXMATRIX_GRP0DST0SRC5CTRL0  94
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC5CTRL1  95
#define EUPHORIA_PFXMATRIX_GRP0DST0SRC5CTRL2  96
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC5CTRL0  97
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC5CTRL1  98
#define EUPHORIA_PFXMATRIX_GRP0DST1SRC5CTRL2  99
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC5CTRL0  100
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC5CTRL1  101
#define EUPHORIA_PFXMATRIX_GRP0DST2SRC5CTRL2  102
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC5CTRL0  103
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC5CTRL1  104
#define EUPHORIA_PFXMATRIX_GRP0DST3SRC5CTRL2  105


//End PolyFX Mod Matrix

/*This is the last control port, + 1 */
#define EUPHORIA_LAST_REGULAR_CONTROL_PORT 106

/*The first port to use when enumerating the ports for mod_matrix controls.  All of the mod_matrix ports should be sequential,
 * any additional ports should prepend this port number*/
#define EUPHORIA_FIRST_SAMPLE_TABLE_PORT 106

/*The range of ports for sample pitch*/
#define EUPHORIA_SAMPLE_PITCH_PORT_RANGE_MIN     EUPHORIA_FIRST_SAMPLE_TABLE_PORT
#define EUPHORIA_SAMPLE_PITCH_PORT_RANGE_MAX     (EUPHORIA_SAMPLE_PITCH_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

/*The range of ports for the low note to play a sample on*/
#define EUPHORIA_PLAY_PITCH_LOW_PORT_RANGE_MIN   EUPHORIA_SAMPLE_PITCH_PORT_RANGE_MAX
#define EUPHORIA_PLAY_PITCH_LOW_PORT_RANGE_MAX   (EUPHORIA_PLAY_PITCH_LOW_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

/*The range of ports for the high note to play a sample on*/
#define EUPHORIA_PLAY_PITCH_HIGH_PORT_RANGE_MIN  EUPHORIA_PLAY_PITCH_LOW_PORT_RANGE_MAX
#define EUPHORIA_PLAY_PITCH_HIGH_PORT_RANGE_MAX  (EUPHORIA_PLAY_PITCH_HIGH_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

/*The range of ports for sample volume*/
#define LMS_SAMPLE_VOLUME_PORT_RANGE_MIN    EUPHORIA_PLAY_PITCH_HIGH_PORT_RANGE_MAX
#define EUPHORIA_SAMPLE_VOLUME_PORT_RANGE_MAX    (LMS_SAMPLE_VOLUME_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

#define EUPHORIA_SAMPLE_START_PORT_RANGE_MIN    EUPHORIA_SAMPLE_VOLUME_PORT_RANGE_MAX
#define EUPHORIA_SAMPLE_START_PORT_RANGE_MAX    (EUPHORIA_SAMPLE_START_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

#define EUPHORIA_SAMPLE_END_PORT_RANGE_MIN    EUPHORIA_SAMPLE_START_PORT_RANGE_MAX
#define EUPHORIA_SAMPLE_END_PORT_RANGE_MAX    (EUPHORIA_SAMPLE_END_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

#define EUPHORIA_SAMPLE_VEL_SENS_PORT_RANGE_MIN    EUPHORIA_SAMPLE_END_PORT_RANGE_MAX
#define EUPHORIA_SAMPLE_VEL_SENS_PORT_RANGE_MAX    (EUPHORIA_SAMPLE_VEL_SENS_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

#define EUPHORIA_SAMPLE_VEL_LOW_PORT_RANGE_MIN    EUPHORIA_SAMPLE_VEL_SENS_PORT_RANGE_MAX
#define EUPHORIA_SAMPLE_VEL_LOW_PORT_RANGE_MAX    (EUPHORIA_SAMPLE_VEL_LOW_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

#define EUPHORIA_SAMPLE_VEL_HIGH_PORT_RANGE_MIN    EUPHORIA_SAMPLE_VEL_LOW_PORT_RANGE_MAX
#define EUPHORIA_SAMPLE_VEL_HIGH_PORT_RANGE_MAX    (EUPHORIA_SAMPLE_VEL_HIGH_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

#define EUPHORIA_PITCH_PORT_RANGE_MIN    EUPHORIA_SAMPLE_VEL_HIGH_PORT_RANGE_MAX
#define EUPHORIA_PITCH_PORT_RANGE_MAX    (EUPHORIA_PITCH_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

#define EUPHORIA_TUNE_PORT_RANGE_MIN    EUPHORIA_PITCH_PORT_RANGE_MAX
#define EUPHORIA_TUNE_PORT_RANGE_MAX    (EUPHORIA_TUNE_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

#define EUPHORIA_SAMPLE_INTERPOLATION_MODE_PORT_RANGE_MIN    EUPHORIA_TUNE_PORT_RANGE_MAX
#define EUPHORIA_SAMPLE_INTERPOLATION_MODE_PORT_RANGE_MAX    (EUPHORIA_SAMPLE_INTERPOLATION_MODE_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

#define EUPHORIA_SAMPLE_LOOP_START_PORT_RANGE_MIN    EUPHORIA_SAMPLE_INTERPOLATION_MODE_PORT_RANGE_MAX
#define EUPHORIA_SAMPLE_LOOP_START_PORT_RANGE_MAX    (EUPHORIA_SAMPLE_LOOP_START_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

#define EUPHORIA_SAMPLE_LOOP_END_PORT_RANGE_MIN    EUPHORIA_SAMPLE_LOOP_START_PORT_RANGE_MAX
#define EUPHORIA_SAMPLE_LOOP_END_PORT_RANGE_MAX    (EUPHORIA_SAMPLE_LOOP_END_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

#define EUPHORIA_SAMPLE_LOOP_MODE_PORT_RANGE_MIN    EUPHORIA_SAMPLE_LOOP_END_PORT_RANGE_MAX
#define EUPHORIA_SAMPLE_LOOP_MODE_PORT_RANGE_MAX    (EUPHORIA_SAMPLE_LOOP_MODE_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

//MonoFX0
#define EUPHORIA_MONO_FX0_KNOB0_PORT_RANGE_MIN    EUPHORIA_SAMPLE_LOOP_MODE_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX0_KNOB0_PORT_RANGE_MAX    (EUPHORIA_MONO_FX0_KNOB0_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)

#define EUPHORIA_MONO_FX0_KNOB1_PORT_RANGE_MIN    EUPHORIA_MONO_FX0_KNOB0_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX0_KNOB1_PORT_RANGE_MAX    (EUPHORIA_MONO_FX0_KNOB1_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)

#define EUPHORIA_MONO_FX0_KNOB2_PORT_RANGE_MIN    EUPHORIA_MONO_FX0_KNOB1_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX0_KNOB2_PORT_RANGE_MAX    (EUPHORIA_MONO_FX0_KNOB2_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)

#define EUPHORIA_MONO_FX0_COMBOBOX_PORT_RANGE_MIN    EUPHORIA_MONO_FX0_KNOB2_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX0_COMBOBOX_PORT_RANGE_MAX    (EUPHORIA_MONO_FX0_COMBOBOX_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)
//MonoFX1
#define EUPHORIA_MONO_FX1_KNOB0_PORT_RANGE_MIN    EUPHORIA_MONO_FX0_COMBOBOX_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX1_KNOB0_PORT_RANGE_MAX    (EUPHORIA_MONO_FX1_KNOB0_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)

#define EUPHORIA_MONO_FX1_KNOB1_PORT_RANGE_MIN    EUPHORIA_MONO_FX1_KNOB0_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX1_KNOB1_PORT_RANGE_MAX    (EUPHORIA_MONO_FX1_KNOB1_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)

#define EUPHORIA_MONO_FX1_KNOB2_PORT_RANGE_MIN    EUPHORIA_MONO_FX1_KNOB1_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX1_KNOB2_PORT_RANGE_MAX    (EUPHORIA_MONO_FX1_KNOB2_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)

#define EUPHORIA_MONO_FX1_COMBOBOX_PORT_RANGE_MIN    EUPHORIA_MONO_FX1_KNOB2_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX1_COMBOBOX_PORT_RANGE_MAX    (EUPHORIA_MONO_FX1_COMBOBOX_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)
//MonoFX2
#define EUPHORIA_MONO_FX2_KNOB0_PORT_RANGE_MIN    EUPHORIA_MONO_FX1_COMBOBOX_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX2_KNOB0_PORT_RANGE_MAX    (EUPHORIA_MONO_FX2_KNOB0_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)

#define EUPHORIA_MONO_FX2_KNOB1_PORT_RANGE_MIN    EUPHORIA_MONO_FX2_KNOB0_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX2_KNOB1_PORT_RANGE_MAX    (EUPHORIA_MONO_FX2_KNOB1_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)

#define EUPHORIA_MONO_FX2_KNOB2_PORT_RANGE_MIN    EUPHORIA_MONO_FX2_KNOB1_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX2_KNOB2_PORT_RANGE_MAX    (EUPHORIA_MONO_FX2_KNOB2_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)

#define EUPHORIA_MONO_FX2_COMBOBOX_PORT_RANGE_MIN    EUPHORIA_MONO_FX2_KNOB2_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX2_COMBOBOX_PORT_RANGE_MAX    (EUPHORIA_MONO_FX2_COMBOBOX_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)
//MonoFX3
#define EUPHORIA_MONO_FX3_KNOB0_PORT_RANGE_MIN    EUPHORIA_MONO_FX2_COMBOBOX_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX3_KNOB0_PORT_RANGE_MAX    (EUPHORIA_MONO_FX3_KNOB0_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)

#define EUPHORIA_MONO_FX3_KNOB1_PORT_RANGE_MIN    EUPHORIA_MONO_FX3_KNOB0_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX3_KNOB1_PORT_RANGE_MAX    (EUPHORIA_MONO_FX3_KNOB1_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)

#define EUPHORIA_MONO_FX3_KNOB2_PORT_RANGE_MIN    EUPHORIA_MONO_FX3_KNOB1_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX3_KNOB2_PORT_RANGE_MAX    (EUPHORIA_MONO_FX3_KNOB2_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)

#define EUPHORIA_MONO_FX3_COMBOBOX_PORT_RANGE_MIN    EUPHORIA_MONO_FX3_KNOB2_PORT_RANGE_MAX
#define EUPHORIA_MONO_FX3_COMBOBOX_PORT_RANGE_MAX    (EUPHORIA_MONO_FX3_COMBOBOX_PORT_RANGE_MIN + EUPHORIA_MONO_FX_GROUPS_COUNT)
//Sample FX Group
#define EUPHORIA_SAMPLE_MONO_FX_GROUP_PORT_RANGE_MIN    EUPHORIA_MONO_FX3_COMBOBOX_PORT_RANGE_MAX
#define EUPHORIA_SAMPLE_MONO_FX_GROUP_PORT_RANGE_MAX    (EUPHORIA_SAMPLE_MONO_FX_GROUP_PORT_RANGE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

//Noise amp
#define EUPHORIA_NOISE_AMP_MIN EUPHORIA_SAMPLE_MONO_FX_GROUP_PORT_RANGE_MAX
#define EUPHORIA_NOISE_AMP_MAX (EUPHORIA_NOISE_AMP_MIN + EUPHORIA_MAX_SAMPLE_COUNT)
//Noise type
#define EUPHORIA_NOISE_TYPE_MIN EUPHORIA_NOISE_AMP_MAX
#define EUPHORIA_NOISE_TYPE_MAX (EUPHORIA_NOISE_TYPE_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

//sample fade-in
#define EUPHORIA_SAMPLE_FADE_IN_MIN EUPHORIA_NOISE_TYPE_MAX
#define EUPHORIA_SAMPLE_FADE_IN_MAX (EUPHORIA_SAMPLE_FADE_IN_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

//sample fade-out
#define EUPHORIA_SAMPLE_FADE_OUT_MIN EUPHORIA_SAMPLE_FADE_IN_MAX
#define EUPHORIA_SAMPLE_FADE_OUT_MAX (EUPHORIA_SAMPLE_FADE_OUT_MIN + EUPHORIA_MAX_SAMPLE_COUNT)

#define EUPHORIA_FIRST_EQ_PORT EUPHORIA_SAMPLE_FADE_OUT_MAX

/* Stacked as:
 * 100 *
 *     [freq, bw, gain] * 6
 */
#define EUPHORIA_LAST_EQ_PORT (EUPHORIA_FIRST_EQ_PORT + (18 * 100))

#define EUPHORIA_LFO_PITCH_FINE EUPHORIA_LAST_EQ_PORT
#define EUPHORIA_MIN_NOTE (EUPHORIA_LFO_PITCH_FINE + 1)
#define EUPHORIA_MAX_NOTE (EUPHORIA_MIN_NOTE + 1)
#define EUPHORIA_MASTER_PITCH (EUPHORIA_MAX_NOTE + 1)
#define EUPHORIA_ADSR_LIN_MAIN (EUPHORIA_MASTER_PITCH + 1)

#define EUPHORIA_PORT_COUNT (EUPHORIA_ADSR_LIN_MAIN + 1)

#endif	/* PORTS_H */

