/*
This file is part of the MusiKernel project, Copyright MusiKernel Team

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*/

#ifndef WAYV_SYNTH_H
#define	WAYV_SYNTH_H

#ifdef	__cplusplus
extern "C" {
#endif

#include "../../include/pydaw_plugin.h"
#include "libmodsynth.h"
#include "../../libmodsynth/lib/voice.h"

//How many modular PolyFX
#define WAYV_MODULAR_POLYFX_COUNT 4
//How many ports per PolyFX, 3 knobs and a combobox
#define WAYV_PORTS_PER_MOD_EFFECT 4
//How many knobs per PolyFX, 3 knobs


#define WAYV_OUTPUT0  0
#define WAYV_OUTPUT1  1

#define WAYV_FIRST_CONTROL_PORT 2
#define WAYV_ATTACK_MAIN  2
#define WAYV_DECAY_MAIN   3
#define WAYV_SUSTAIN_MAIN 4
#define WAYV_RELEASE_MAIN 5
#define WAYV_NOISE_AMP 6
#define WAYV_OSC1_TYPE 7
#define WAYV_OSC1_PITCH 8
#define WAYV_OSC1_TUNE 9
#define WAYV_OSC1_VOLUME 10
#define WAYV_OSC2_TYPE 11
#define WAYV_OSC2_PITCH 12
#define WAYV_OSC2_TUNE 13
#define WAYV_OSC2_VOLUME 14
#define WAYV_MASTER_VOLUME 15
#define WAYV_OSC1_UNISON_VOICES 16
#define WAYV_OSC1_UNISON_SPREAD 17
#define WAYV_MASTER_GLIDE 18
#define WAYV_MASTER_PITCHBEND_AMT 19
#define WAYV_ATTACK1  20
#define WAYV_DECAY1   21
#define WAYV_SUSTAIN1 22
#define WAYV_RELEASE1 23
#define WAYV_ATTACK2  24
#define WAYV_DECAY2   25
#define WAYV_SUSTAIN2 26
#define WAYV_RELEASE2 27
#define WAYV_ATTACK_PFX1  28
#define WAYV_DECAY_PFX1   29
#define WAYV_SUSTAIN_PFX1 30
#define WAYV_RELEASE_PFX1 31
#define WAYV_ATTACK_PFX2  32
#define WAYV_DECAY_PFX2   33
#define WAYV_SUSTAIN_PFX2 34
#define WAYV_RELEASE_PFX2 35
#define WAYV_NOISE_TYPE 36
#define WAYV_RAMP_ENV_TIME 37
#define WAYV_LFO_FREQ 38
#define WAYV_LFO_TYPE 39
#define WAYV_FX0_KNOB0  40
#define WAYV_FX0_KNOB1 41
#define WAYV_FX0_KNOB2  42
#define WAYV_FX0_COMBOBOX 43
#define WAYV_FX1_KNOB0  44
#define WAYV_FX1_KNOB1  45
#define WAYV_FX1_KNOB2  46
#define WAYV_FX1_COMBOBOX 47
#define WAYV_FX2_KNOB0  48
#define WAYV_FX2_KNOB1  49
#define WAYV_FX2_KNOB2  50
#define WAYV_FX2_COMBOBOX 51
#define WAYV_FX3_KNOB0  52
#define WAYV_FX3_KNOB1  53
#define WAYV_FX3_KNOB2  54
#define WAYV_FX3_COMBOBOX 55
//PolyFX Mod Matrix
#define WAVV_PFXMATRIX_FIRST_PORT 56

#define WAVV_PFXMATRIX_GRP0DST0SRC0CTRL0  56
#define WAVV_PFXMATRIX_GRP0DST0SRC0CTRL1  57
#define WAVV_PFXMATRIX_GRP0DST0SRC0CTRL2  58
#define WAVV_PFXMATRIX_GRP0DST0SRC1CTRL0  59
#define WAVV_PFXMATRIX_GRP0DST0SRC1CTRL1  60
#define WAVV_PFXMATRIX_GRP0DST0SRC1CTRL2  61
#define WAVV_PFXMATRIX_GRP0DST0SRC2CTRL0  62
#define WAVV_PFXMATRIX_GRP0DST0SRC2CTRL1  63
#define WAVV_PFXMATRIX_GRP0DST0SRC2CTRL2  64
#define WAVV_PFXMATRIX_GRP0DST0SRC3CTRL0  65
#define WAVV_PFXMATRIX_GRP0DST0SRC3CTRL1  66
#define WAVV_PFXMATRIX_GRP0DST0SRC3CTRL2  67
#define WAVV_PFXMATRIX_GRP0DST1SRC0CTRL0  68
#define WAVV_PFXMATRIX_GRP0DST1SRC0CTRL1  69
#define WAVV_PFXMATRIX_GRP0DST1SRC0CTRL2  70
#define WAVV_PFXMATRIX_GRP0DST1SRC1CTRL0  71
#define WAVV_PFXMATRIX_GRP0DST1SRC1CTRL1  72
#define WAVV_PFXMATRIX_GRP0DST1SRC1CTRL2  73
#define WAVV_PFXMATRIX_GRP0DST1SRC2CTRL0  74
#define WAVV_PFXMATRIX_GRP0DST1SRC2CTRL1  75
#define WAVV_PFXMATRIX_GRP0DST1SRC2CTRL2  76
#define WAVV_PFXMATRIX_GRP0DST1SRC3CTRL0  77
#define WAVV_PFXMATRIX_GRP0DST1SRC3CTRL1  78
#define WAVV_PFXMATRIX_GRP0DST1SRC3CTRL2  79
#define WAVV_PFXMATRIX_GRP0DST2SRC0CTRL0  80
#define WAVV_PFXMATRIX_GRP0DST2SRC0CTRL1  81
#define WAVV_PFXMATRIX_GRP0DST2SRC0CTRL2  82
#define WAVV_PFXMATRIX_GRP0DST2SRC1CTRL0  83
#define WAVV_PFXMATRIX_GRP0DST2SRC1CTRL1  84
#define WAVV_PFXMATRIX_GRP0DST2SRC1CTRL2  85
#define WAVV_PFXMATRIX_GRP0DST2SRC2CTRL0  86
#define WAVV_PFXMATRIX_GRP0DST2SRC2CTRL1  87
#define WAVV_PFXMATRIX_GRP0DST2SRC2CTRL2  88
#define WAVV_PFXMATRIX_GRP0DST2SRC3CTRL0  89
#define WAVV_PFXMATRIX_GRP0DST2SRC3CTRL1  90
#define WAVV_PFXMATRIX_GRP0DST2SRC3CTRL2  91
#define WAVV_PFXMATRIX_GRP0DST3SRC0CTRL0  92
#define WAVV_PFXMATRIX_GRP0DST3SRC0CTRL1  93
#define WAVV_PFXMATRIX_GRP0DST3SRC0CTRL2  94
#define WAVV_PFXMATRIX_GRP0DST3SRC1CTRL0  95
#define WAVV_PFXMATRIX_GRP0DST3SRC1CTRL1  96
#define WAVV_PFXMATRIX_GRP0DST3SRC1CTRL2  97
#define WAVV_PFXMATRIX_GRP0DST3SRC2CTRL0  98
#define WAVV_PFXMATRIX_GRP0DST3SRC2CTRL1  99
#define WAVV_PFXMATRIX_GRP0DST3SRC2CTRL2  100
#define WAVV_PFXMATRIX_GRP0DST3SRC3CTRL0  101
#define WAVV_PFXMATRIX_GRP0DST3SRC3CTRL1  102
#define WAVV_PFXMATRIX_GRP0DST3SRC3CTRL2  103

//End PolyFX Mod Matrix

#define WAYV_ADSR1_CHECKBOX 105
#define WAYV_ADSR2_CHECKBOX 106
#define WAYV_LFO_AMP 107
#define WAYV_LFO_PITCH 108
#define WAYV_PITCH_ENV_AMT 109
#define WAYV_OSC2_UNISON_VOICES 110
#define WAYV_OSC2_UNISON_SPREAD 111
#define WAYV_LFO_AMOUNT 112
#define WAYV_OSC3_TYPE 113
#define WAYV_OSC3_PITCH 114
#define WAYV_OSC3_TUNE 115
#define WAYV_OSC3_VOLUME 116
#define WAYV_OSC3_UNISON_VOICES 117
#define WAYV_OSC3_UNISON_SPREAD 118
#define WAYV_OSC1_FM1 119
#define WAYV_OSC1_FM2 120
#define WAYV_OSC1_FM3 121
#define WAYV_OSC2_FM1 122
#define WAYV_OSC2_FM2 123
#define WAYV_OSC2_FM3 124
#define WAYV_OSC3_FM1 125
#define WAYV_OSC3_FM2 126
#define WAYV_OSC3_FM3 127
#define WAYV_ATTACK3  128
#define WAYV_DECAY3   129
#define WAYV_SUSTAIN3 130
#define WAYV_RELEASE3 131
#define WAYV_ADSR3_CHECKBOX 132

#define WAVV_PFXMATRIX_GRP0DST0SRC4CTRL0  133
#define WAVV_PFXMATRIX_GRP0DST0SRC4CTRL1  134
#define WAVV_PFXMATRIX_GRP0DST0SRC4CTRL2  135
#define WAVV_PFXMATRIX_GRP0DST1SRC4CTRL0  136
#define WAVV_PFXMATRIX_GRP0DST1SRC4CTRL1  137
#define WAVV_PFXMATRIX_GRP0DST1SRC4CTRL2  138
#define WAVV_PFXMATRIX_GRP0DST2SRC4CTRL0  139
#define WAVV_PFXMATRIX_GRP0DST2SRC4CTRL1  140
#define WAVV_PFXMATRIX_GRP0DST2SRC4CTRL2  141
#define WAVV_PFXMATRIX_GRP0DST3SRC4CTRL0  142
#define WAVV_PFXMATRIX_GRP0DST3SRC4CTRL1  143
#define WAVV_PFXMATRIX_GRP0DST3SRC4CTRL2  144

#define WAVV_PFXMATRIX_GRP0DST0SRC5CTRL0  145
#define WAVV_PFXMATRIX_GRP0DST0SRC5CTRL1  146
#define WAVV_PFXMATRIX_GRP0DST0SRC5CTRL2  147
#define WAVV_PFXMATRIX_GRP0DST1SRC5CTRL0  148
#define WAVV_PFXMATRIX_GRP0DST1SRC5CTRL1  149
#define WAVV_PFXMATRIX_GRP0DST1SRC5CTRL2  150
#define WAVV_PFXMATRIX_GRP0DST2SRC5CTRL0  151
#define WAVV_PFXMATRIX_GRP0DST2SRC5CTRL1  152
#define WAVV_PFXMATRIX_GRP0DST2SRC5CTRL2  153
#define WAVV_PFXMATRIX_GRP0DST3SRC5CTRL0  154
#define WAVV_PFXMATRIX_GRP0DST3SRC5CTRL1  155
#define WAVV_PFXMATRIX_GRP0DST3SRC5CTRL2  156

#define WAYV_PERC_ENV_TIME1 157
#define WAYV_PERC_ENV_PITCH1 158
#define WAYV_PERC_ENV_TIME2 159
#define WAYV_PERC_ENV_PITCH2 160
#define WAYV_PERC_ENV_ON 161
#define WAYV_RAMP_CURVE 162
#define WAYV_MONO_MODE 163

#define WAYV_OSC4_TYPE 164
#define WAYV_OSC4_PITCH 165
#define WAYV_OSC4_TUNE 166
#define WAYV_OSC4_VOLUME 167
#define WAYV_OSC4_UNISON_VOICES 168
#define WAYV_OSC4_UNISON_SPREAD 169
#define WAYV_OSC1_FM4 170
#define WAYV_OSC2_FM4 171
#define WAYV_OSC3_FM4 172
#define WAYV_OSC4_FM1 173
#define WAYV_OSC4_FM2 174
#define WAYV_OSC4_FM3 175
#define WAYV_OSC4_FM4 176
#define WAYV_ATTACK4  177
#define WAYV_DECAY4   178
#define WAYV_SUSTAIN4 179
#define WAYV_RELEASE4 180
#define WAYV_ADSR4_CHECKBOX 181

#define WAYV_FM_MACRO1 182
#define WAYV_FM_MACRO1_OSC1_FM1 183
#define WAYV_FM_MACRO1_OSC1_FM2 184
#define WAYV_FM_MACRO1_OSC1_FM3 185
#define WAYV_FM_MACRO1_OSC1_FM4 186
#define WAYV_FM_MACRO1_OSC2_FM1 187
#define WAYV_FM_MACRO1_OSC2_FM2 188
#define WAYV_FM_MACRO1_OSC2_FM3 189
#define WAYV_FM_MACRO1_OSC2_FM4 190
#define WAYV_FM_MACRO1_OSC3_FM1 191
#define WAYV_FM_MACRO1_OSC3_FM2 192
#define WAYV_FM_MACRO1_OSC3_FM3 193
#define WAYV_FM_MACRO1_OSC3_FM4 194
#define WAYV_FM_MACRO1_OSC4_FM1 195
#define WAYV_FM_MACRO1_OSC4_FM2 196
#define WAYV_FM_MACRO1_OSC4_FM3 197
#define WAYV_FM_MACRO1_OSC4_FM4 198

#define WAYV_FM_MACRO2 199
#define WAYV_FM_MACRO2_OSC1_FM1 200
#define WAYV_FM_MACRO2_OSC1_FM2 201
#define WAYV_FM_MACRO2_OSC1_FM3 202
#define WAYV_FM_MACRO2_OSC1_FM4 203
#define WAYV_FM_MACRO2_OSC2_FM1 204
#define WAYV_FM_MACRO2_OSC2_FM2 205
#define WAYV_FM_MACRO2_OSC2_FM3 206
#define WAYV_FM_MACRO2_OSC2_FM4 207
#define WAYV_FM_MACRO2_OSC3_FM1 208
#define WAYV_FM_MACRO2_OSC3_FM2 209
#define WAYV_FM_MACRO2_OSC3_FM3 210
#define WAYV_FM_MACRO2_OSC3_FM4 211
#define WAYV_FM_MACRO2_OSC4_FM1 212
#define WAYV_FM_MACRO2_OSC4_FM2 213
#define WAYV_FM_MACRO2_OSC4_FM3 214
#define WAYV_FM_MACRO2_OSC4_FM4 215

#define WAYV_LFO_PHASE 216

#define WAYV_FM_MACRO1_OSC1_VOL 217
#define WAYV_FM_MACRO1_OSC2_VOL 218
#define WAYV_FM_MACRO1_OSC3_VOL 219
#define WAYV_FM_MACRO1_OSC4_VOL 220

#define WAYV_FM_MACRO2_OSC1_VOL 221
#define WAYV_FM_MACRO2_OSC2_VOL 222
#define WAYV_FM_MACRO2_OSC3_VOL 223
#define WAYV_FM_MACRO2_OSC4_VOL 224
#define WAYV_LFO_PITCH_FINE 225
#define WAYV_ADSR_PREFX 226

#define WAYV_ADSR1_DELAY 227
#define WAYV_ADSR2_DELAY 228
#define WAYV_ADSR3_DELAY 229
#define WAYV_ADSR4_DELAY 230

#define WAYV_ADSR1_HOLD 231
#define WAYV_ADSR2_HOLD 232
#define WAYV_ADSR3_HOLD 233
#define WAYV_ADSR4_HOLD 234

#define WAYV_PFX_ADSR_DELAY 235
#define WAYV_PFX_ADSR_F_DELAY 236

#define WAYV_PFX_ADSR_HOLD 237
#define WAYV_PFX_ADSR_F_HOLD 238

#define WAYV_HOLD_MAIN  239

#define WAYV_DELAY_NOISE  240
#define WAYV_ATTACK_NOISE  241
#define WAYV_HOLD_NOISE  242
#define WAYV_DECAY_NOISE   243
#define WAYV_SUSTAIN_NOISE 244
#define WAYV_RELEASE_NOISE 245
#define WAYV_ADSR_NOISE_ON 246

#define WAYV_DELAY_LFO  247
#define WAYV_ATTACK_LFO  248
#define WAYV_HOLD_LFO  249
#define WAYV_DECAY_LFO   250
#define WAYV_SUSTAIN_LFO 251
#define WAYV_RELEASE_LFO 252
#define WAYV_ADSR_LFO_ON 253

#define WAYV_OSC5_TYPE 254
#define WAYV_OSC5_PITCH 255
#define WAYV_OSC5_TUNE 256
#define WAYV_OSC5_VOLUME 257
#define WAYV_OSC5_UNISON_VOICES 258
#define WAYV_OSC5_UNISON_SPREAD 259
#define WAYV_OSC1_FM5 260
#define WAYV_OSC2_FM5 261
#define WAYV_OSC3_FM5 262
#define WAYV_OSC4_FM5 263
#define WAYV_OSC5_FM5 264
#define WAYV_OSC6_FM5 265
#define WAYV_ADSR5_DELAY 266
#define WAYV_ATTACK5  267
#define WAYV_ADSR5_HOLD 268
#define WAYV_DECAY5   269
#define WAYV_SUSTAIN5 270
#define WAYV_RELEASE5 271
#define WAYV_ADSR5_CHECKBOX 272

#define WAYV_OSC6_TYPE 273
#define WAYV_OSC6_PITCH 274
#define WAYV_OSC6_TUNE 275
#define WAYV_OSC6_VOLUME 276
#define WAYV_OSC6_UNISON_VOICES 277
#define WAYV_OSC6_UNISON_SPREAD 278
#define WAYV_OSC1_FM6 279
#define WAYV_OSC2_FM6 280
#define WAYV_OSC3_FM6 281
#define WAYV_OSC4_FM6 282
#define WAYV_OSC5_FM6 283
#define WAYV_OSC6_FM6 284
#define WAYV_ADSR6_DELAY 285
#define WAYV_ATTACK6 286
#define WAYV_ADSR6_HOLD 287
#define WAYV_DECAY6   288
#define WAYV_SUSTAIN6 289
#define WAYV_RELEASE6 290
#define WAYV_ADSR6_CHECKBOX 291


#define WAYV_FM_MACRO1_OSC1_FM5 292
#define WAYV_FM_MACRO1_OSC2_FM5 293
#define WAYV_FM_MACRO1_OSC3_FM5 294
#define WAYV_FM_MACRO1_OSC4_FM5 295
#define WAYV_FM_MACRO1_OSC5_FM5 296
#define WAYV_FM_MACRO1_OSC6_FM5 297

#define WAYV_FM_MACRO1_OSC1_FM6 298
#define WAYV_FM_MACRO1_OSC2_FM6 299
#define WAYV_FM_MACRO1_OSC3_FM6 300
#define WAYV_FM_MACRO1_OSC4_FM6 301
#define WAYV_FM_MACRO1_OSC5_FM6 302
#define WAYV_FM_MACRO1_OSC6_FM6 303

#define WAYV_FM_MACRO1_OSC5_FM1 304
#define WAYV_FM_MACRO1_OSC5_FM2 305
#define WAYV_FM_MACRO1_OSC5_FM3 306
#define WAYV_FM_MACRO1_OSC5_FM4 307

#define WAYV_FM_MACRO1_OSC6_FM1 308
#define WAYV_FM_MACRO1_OSC6_FM2 309
#define WAYV_FM_MACRO1_OSC6_FM3 310
#define WAYV_FM_MACRO1_OSC6_FM4 311

#define WAYV_FM_MACRO1_OSC5_VOL 312
#define WAYV_FM_MACRO1_OSC6_VOL 313

#define WAYV_FM_MACRO2_OSC1_FM5 314
#define WAYV_FM_MACRO2_OSC2_FM5 315
#define WAYV_FM_MACRO2_OSC3_FM5 316
#define WAYV_FM_MACRO2_OSC4_FM5 317
#define WAYV_FM_MACRO2_OSC5_FM5 318
#define WAYV_FM_MACRO2_OSC6_FM5 319

#define WAYV_FM_MACRO2_OSC1_FM6 320
#define WAYV_FM_MACRO2_OSC2_FM6 321
#define WAYV_FM_MACRO2_OSC3_FM6 322
#define WAYV_FM_MACRO2_OSC4_FM6 323
#define WAYV_FM_MACRO2_OSC5_FM6 324
#define WAYV_FM_MACRO2_OSC6_FM6 325

#define WAYV_FM_MACRO2_OSC5_FM1 326
#define WAYV_FM_MACRO2_OSC5_FM2 327
#define WAYV_FM_MACRO2_OSC5_FM3 328
#define WAYV_FM_MACRO2_OSC5_FM4 329

#define WAYV_FM_MACRO2_OSC6_FM1 330
#define WAYV_FM_MACRO2_OSC6_FM2 331
#define WAYV_FM_MACRO2_OSC6_FM3 332
#define WAYV_FM_MACRO2_OSC6_FM4 333

#define WAYV_FM_MACRO2_OSC5_VOL 334
#define WAYV_FM_MACRO2_OSC6_VOL 335

#define WAYV_OSC5_FM1 336
#define WAYV_OSC5_FM2 337
#define WAYV_OSC5_FM3 338
#define WAYV_OSC5_FM4 339

#define WAYV_OSC6_FM1 340
#define WAYV_OSC6_FM2 341
#define WAYV_OSC6_FM3 342
#define WAYV_OSC6_FM4 343
#define WAYV_NOISE_PREFX 344

#define WAVV_PFXMATRIX_GRP0DST0SRC6CTRL0 345
#define WAVV_PFXMATRIX_GRP0DST0SRC6CTRL1 346
#define WAVV_PFXMATRIX_GRP0DST0SRC6CTRL2 347
#define WAVV_PFXMATRIX_GRP0DST1SRC6CTRL0 348
#define WAVV_PFXMATRIX_GRP0DST1SRC6CTRL1 349
#define WAVV_PFXMATRIX_GRP0DST1SRC6CTRL2 350
#define WAVV_PFXMATRIX_GRP0DST2SRC6CTRL0 351
#define WAVV_PFXMATRIX_GRP0DST2SRC6CTRL1 352
#define WAVV_PFXMATRIX_GRP0DST2SRC6CTRL2 353
#define WAVV_PFXMATRIX_GRP0DST3SRC6CTRL0 354
#define WAVV_PFXMATRIX_GRP0DST3SRC6CTRL1 355
#define WAVV_PFXMATRIX_GRP0DST3SRC6CTRL2 356

#define WAVV_PFXMATRIX_GRP0DST0SRC7CTRL0 357
#define WAVV_PFXMATRIX_GRP0DST0SRC7CTRL1 358
#define WAVV_PFXMATRIX_GRP0DST0SRC7CTRL2 359
#define WAVV_PFXMATRIX_GRP0DST1SRC7CTRL0 360
#define WAVV_PFXMATRIX_GRP0DST1SRC7CTRL1 361
#define WAVV_PFXMATRIX_GRP0DST1SRC7CTRL2 362
#define WAVV_PFXMATRIX_GRP0DST2SRC7CTRL0 363
#define WAVV_PFXMATRIX_GRP0DST2SRC7CTRL1 364
#define WAVV_PFXMATRIX_GRP0DST2SRC7CTRL2 365
#define WAVV_PFXMATRIX_GRP0DST3SRC7CTRL0 366
#define WAVV_PFXMATRIX_GRP0DST3SRC7CTRL1 367
#define WAVV_PFXMATRIX_GRP0DST3SRC7CTRL2 368

#define WAYV_MIN_NOTE 369
#define WAYV_MAX_NOTE 370
#define WAYV_MASTER_PITCH 371
#define WAYV_ADSR_LIN_MAIN 372

/* must be 1 + highest value above
 * CHANGE THIS IF YOU ADD OR TAKE AWAY ANYTHING*/
#define WAYV_COUNT 373

#define WAYV_POLYPHONY   16
#define WAYV_POLYPHONY_THRESH 12

typedef struct
{
    PYFX_Data *output0;
    PYFX_Data *output1;
    PYFX_Data *adsr_lin_main;
    PYFX_Data *attack_main;
    PYFX_Data *hold_main;
    PYFX_Data *decay_main;
    PYFX_Data *sustain_main;
    PYFX_Data *release_main;

    PYFX_Data *attack[WAYV_OSC_COUNT];
    PYFX_Data *decay[WAYV_OSC_COUNT];
    PYFX_Data *sustain[WAYV_OSC_COUNT];
    PYFX_Data *release[WAYV_OSC_COUNT];

    PYFX_Data *adsr_checked[WAYV_OSC_COUNT];

    PYFX_Data *adsr_fm_delay[WAYV_OSC_COUNT];
    PYFX_Data *adsr_fm_hold[WAYV_OSC_COUNT];

    PYFX_Data *osc_pitch[WAYV_OSC_COUNT];
    PYFX_Data *osc_tune[WAYV_OSC_COUNT];
    PYFX_Data *osc_vol[WAYV_OSC_COUNT];
    PYFX_Data *osc_type[WAYV_OSC_COUNT];

    PYFX_Data *osc_fm[WAYV_OSC_COUNT][WAYV_OSC_COUNT];

    PYFX_Data *master_vol;

    PYFX_Data *pfx_delay;
    PYFX_Data *pfx_attack;
    PYFX_Data *pfx_hold;
    PYFX_Data *pfx_decay;
    PYFX_Data *pfx_sustain;
    PYFX_Data *pfx_release;

    PYFX_Data *pfx_delay_f;
    PYFX_Data *pfx_attack_f;
    PYFX_Data *pfx_hold_f;
    PYFX_Data *pfx_decay_f;
    PYFX_Data *pfx_sustain_f;
    PYFX_Data *pfx_release_f;

    PYFX_Data *noise_amp;
    PYFX_Data *noise_type;

    PYFX_Data *noise_adsr_on;
    PYFX_Data *noise_prefx;
    PYFX_Data *noise_delay;
    PYFX_Data *noise_attack;
    PYFX_Data *noise_hold;
    PYFX_Data *noise_decay;
    PYFX_Data *noise_sustain;
    PYFX_Data *noise_release;

    PYFX_Data *osc_uni_voice[WAYV_OSC_COUNT];
    PYFX_Data *osc_uni_spread[WAYV_OSC_COUNT];

    PYFX_Data *master_glide;
    PYFX_Data *master_pb_amt;

    PYFX_Data *pitch_env_time;
    PYFX_Data *pitch_env_amt;
    PYFX_Data *ramp_curve;

    PYFX_Data *lfo_freq;
    PYFX_Data *lfo_type;
    PYFX_Data *lfo_phase;

    PYFX_Data *lfo_amp;
    PYFX_Data *lfo_pitch;
    PYFX_Data *lfo_pitch_fine;
    PYFX_Data *lfo_amount;

    PYFX_Data *lfo_adsr_on;
    PYFX_Data *lfo_delay;
    PYFX_Data *lfo_attack;
    PYFX_Data *lfo_hold;
    PYFX_Data *lfo_decay;
    PYFX_Data *lfo_sustain;
    PYFX_Data *lfo_release;

    PYFX_Data *perc_env_time1;
    PYFX_Data *perc_env_pitch1;
    PYFX_Data *perc_env_time2;
    PYFX_Data *perc_env_pitch2;
    PYFX_Data *perc_env_on;
    PYFX_Data *adsr_prefx;

    PYFX_Data *fm_macro[2];
    PYFX_Data *fm_macro_values[2][WAYV_OSC_COUNT][WAYV_OSC_COUNT];
    PYFX_Data *amp_macro_values[2][WAYV_OSC_COUNT];

    PYFX_Data *mono_mode;

    PYFX_Data *min_note;
    PYFX_Data *max_note;
    PYFX_Data *master_pitch;

    //Corresponds to the actual knobs on the effects themselves,
    //not the mod matrix
    PYFX_Data *pfx_mod_knob
        [WAYV_MODULAR_POLYFX_COUNT][WAYV_CONTROLS_PER_MOD_EFFECT];

    PYFX_Data *fx_combobox[WAYV_MODULAR_POLYFX_COUNT];

    //PolyFX Mod Matrix
    //Corresponds to the mod matrix spinboxes
    PYFX_Data *polyfx_mod_matrix
            [WAYV_MODULAR_POLYFX_COUNT][WAYV_MODULATOR_COUNT]
            [WAYV_CONTROLS_PER_MOD_EFFECT];

    t_wayv_poly_voice * data[WAYV_POLYPHONY];

    t_voc_voices * voices;

    long         sampleNo;

    float fs;
    t_wayv_mono_modules * mono_modules;

    float sv_last_note;  //For glide
    float sv_pitch_bend_value;
    t_plugin_event_queue midi_queue;
    float * port_table;
    t_plugin_event_queue atm_queue;
    t_plugin_cc_map cc_map;
    PYFX_Descriptor * descriptor;
} t_wayv;




#ifdef	__cplusplus
}
#endif

#endif	/* WAYV_SYNTH_H */

